@@ -1,4 +1,5 @@
 {-# LANGUAGE RankNTypes #-}
+{-# LANGUAGE DeriveFunctor #-}
 -- | If this is your first time with conduit, you should probably start with
 -- the tutorial:
 -- <https://haskell.fpcomplete.com/user/snoyberg/library-documentation/conduit-overview>.
@@ -15,6 +16,10 @@ module Data.Conduit
     , (=$)
     , (=$=)
 
+      -- *** Fuse with upstream results
+    , fuseBoth
+    , fuseUpstream
+
       -- ** Primitives
     , await
     , yield
@@ -42,13 +47,29 @@ module Data.Conduit
     , mapOutput
     , mapOutputMaybe
     , mapInput
+    , passthroughSink
 
       -- * Connect-and-resume
     , ResumableSource
+    , newResumableSource
     , ($$+)
     , ($$++)
     , ($$+-)
+    , ($=+)
     , unwrapResumable
+    , closeResumableSource
+
+      -- ** For @Conduit@s
+    , ResumableConduit
+    , newResumableConduit
+    , (=$$+)
+    , (=$$++)
+    , (=$$+-)
+    , unwrapResumableConduit
+
+      -- * Fusion with leftovers
+    , fuseLeftovers
+    , fuseReturnLeftovers
 
       -- * Flushing
     , Flush (..)
@@ -62,26 +83,19 @@ module Data.Conduit
     , ZipSink (..)
     , sequenceSinks
 
-      -- * Convenience re-exports
-    , ResourceT
-    , MonadResource
-    , MonadThrow (..)
-    , MonadUnsafeIO (..)
-    , runResourceT
-    , ExceptionT (..)
-    , runExceptionT_
-    , runException
-    , runException_
-    , MonadBaseControl
+      -- ** ZipConduit
+    , ZipConduit (..)
+    , sequenceConduits
     ) where
 
-import Control.Monad.Trans.Resource
-import Data.Conduit.Internal hiding (await, awaitForever, yield, yieldOr, leftover, bracketP, addCleanup, transPipe, mapOutput, mapOutputMaybe, mapInput)
+import Control.Monad.Trans.Class (lift)
+import Data.Conduit.Internal hiding (await, awaitForever, yield, yieldOr, leftover, bracketP, addCleanup, transPipe, mapOutput, mapOutputMaybe, mapInput, yieldM)
 import qualified Data.Conduit.Internal as CI
 import Control.Monad.Morph (hoist)
-import Control.Monad (liftM, forever)
+import Control.Monad (liftM, forever, when, unless)
 import Control.Applicative (Applicative (..))
 import Data.Traversable (Traversable (..))
+import Control.Monad.Trans.Resource (MonadResource)
 
 -- Define fixity of all our operators
 infixr 0 $$
@@ -91,6 +105,7 @@ infixr 2 =$=
 infixr 0 $$+
 infixr 0 $$++
 infixr 0 $$+-
+infixl 1 $=+
 
 -- | The connect operator, which pulls data from a source and pushes to a sink.
 -- If you would like to keep the @Source@ open to be used for other
@@ -111,8 +126,11 @@ src $$ sink = do
 --
 -- Leftover data from the @Conduit@ will be discarded.
 --
+-- Note: Since version 1.0.18, this operator has been generalized to be
+-- identical to @=$=@.
+--
 -- Since 0.4.0
-($=) :: Monad m => Source m a -> Conduit a m b -> Source m b
+($=) :: Monad m => Conduit a m b -> ConduitM b c m r -> ConduitM a c m r
 ConduitM src $= ConduitM con = ConduitM $ pipeL src con
 {-# INLINE ($=) #-}
 
@@ -123,8 +141,11 @@ ConduitM src $= ConduitM con = ConduitM $ pipeL src con
 --
 -- Leftover data returned from the @Sink@ will be discarded.
 --
+-- Note: Since version 1.0.18, this operator has been generalized to be
+-- identical to @=$=@.
+--
 -- Since 0.4.0
-(=$) :: Monad m => Conduit a m b -> Sink b m c -> Sink a m c
+(=$) :: Monad m => Conduit a m b -> ConduitM b c m r -> ConduitM a c m r
 ConduitM con =$ ConduitM sink = ConduitM $ pipeL con sink
 {-# INLINE (=$) #-}
 
@@ -145,6 +166,8 @@ ConduitM left =$= ConduitM right = ConduitM $ pipeL left right
 -- Since 0.5.0
 await :: Monad m => Consumer i m (Maybe i)
 await = ConduitM CI.await
+{-# RULES "await >>= maybe" forall x y. await >>= maybe x y = ConduitM (NeedInput (unConduitM . y) (unConduitM . const x)) #-}
+{-# INLINE [1] await #-}
 
 -- | Send a value downstream to the next component to consume. If the
 -- downstream component terminates, this call will never return control. If you
@@ -155,6 +178,23 @@ yield :: Monad m
       => o -- ^ output value
       -> ConduitM i o m ()
 yield = ConduitM . CI.yield
+{-# INLINE [1] yield #-}
+
+yieldM :: Monad m => m o -> ConduitM i o m ()
+yieldM = ConduitM . CI.yieldM
+{-# INLINE [1] yieldM #-}
+
+{-# RULES
+    "yield o >> p" forall o (p :: ConduitM i o m r). yield o >> p = ConduitM (HaveOutput (unConduitM p) (return ()) o)
+  ; "mapM_ yield" mapM_ yield = ConduitM . sourceList
+  ; "yieldOr o c >> p" forall o c (p :: ConduitM i o m r). yieldOr o c >> p =
+        ConduitM (HaveOutput (unConduitM p) c o)
+  ; "when yield next" forall b o p. when b (yield o) >> p =
+        if b then ConduitM (HaveOutput (unConduitM p) (return ()) o) else p
+  ; "unless yield next" forall b o p. unless b (yield o) >> p =
+        if b then p else ConduitM (HaveOutput (unConduitM p) (return ()) o)
+  ; "lift m >>= yield" forall m. lift m >>= yield = yieldM m
+   #-}
 
 -- | Provide a single piece of leftover input to be consumed by the next
 -- component in the current monadic binding.
@@ -165,6 +205,9 @@ yield = ConduitM . CI.yield
 -- Since 0.5.0
 leftover :: i -> ConduitM i o m ()
 leftover = ConduitM . CI.leftover
+{-# INLINE [1] leftover #-}
+{-# RULES "leftover l >> p" forall l (p :: ConduitM i o m r). leftover l >> p =
+    ConduitM (Leftover (unConduitM p) l) #-}
 
 -- | Perform some allocation and run an inner component. Two guarantees are
 -- given about resource finalization:
@@ -207,6 +250,7 @@ yieldOr :: Monad m
         -> m () -- ^ finalizer
         -> ConduitM i o m ()
 yieldOr o m = ConduitM $ CI.yieldOr o m
+{-# INLINE [1] yieldOr #-}
 
 -- | Wait for input forever, calling the given inner component for each piece of
 -- new input. Returns the upstream result type.
@@ -289,6 +333,21 @@ rsrc $$+- sink = do
     return res
 {-# INLINE ($$+-) #-}
 
+-- | Left fusion for a resumable source.
+--
+-- Since 1.0.16
+($=+) :: Monad m => ResumableSource m a -> Conduit a m b -> ResumableSource m b
+ResumableSource src final $=+ sink = ResumableSource (src $= sink) final
+
+-- | Execute the finalizer associated with a @ResumableSource@, rendering the
+-- @ResumableSource@ invalid for further use.
+--
+-- This is just a more explicit version of @$$+- return ()@.
+--
+-- Since 1.1.3
+closeResumableSource :: Monad m => ResumableSource m a -> m ()
+closeResumableSource = ($$+- return ())
+
 -- | Provide for a stream of data that can be flushed.
 --
 -- A number of @Conduit@s (e.g., zlib compression) need the ability to flush
@@ -302,8 +361,8 @@ instance Functor Flush where
     fmap _ Flush = Flush
     fmap f (Chunk a) = Chunk (f a)
 
--- | A wrapper for defining an 'Applicative' instance for 'Sink's which allows
--- to combine sinks together, generalizing 'zipSources'. A combined sources
+-- | A wrapper for defining an 'Applicative' instance for 'Source's which allows
+-- to combine sources together, generalizing 'zipSources'. A combined source
 -- will take input yielded from each of its @Source@s until any of them stop
 -- producing output.
 --
@@ -316,7 +375,7 @@ instance Monad m => Applicative (ZipSource m) where
     pure  = ZipSource . forever . yield
     (ZipSource f) <*> (ZipSource x) = ZipSource $ zipSourcesApp f x
 
--- | Coalesce all values yielding by all of the @Source@s.
+-- | Coalesce all values yielded by all of the @Source@s.
 --
 -- Implemented on top of @ZipSource@, see that data type for more details.
 --
@@ -357,3 +416,98 @@ instance Monad m => Applicative (ZipSink i m) where
 -- Since 1.0.13
 sequenceSinks :: (Traversable f, Monad m) => f (Sink i m r) -> Sink i m (f r)
 sequenceSinks = getZipSink . sequenceA . fmap ZipSink
+
+-- | The connect-and-resume operator. This does not close the @Conduit@, but
+-- instead returns it to be used again. This allows a @Conduit@ to be used
+-- incrementally in a large program, without forcing the entire program to live
+-- in the @Sink@ monad.
+--
+-- Leftover data returned from the @Sink@ will be discarded.
+--
+-- Mnemonic: connect + do more.
+--
+-- Since 1.0.17
+(=$$+) :: Monad m => Conduit a m b -> Sink b m r -> Sink a m (ResumableConduit a m b, r)
+(=$$+) conduit = connectResumeConduit (ResumableConduit conduit (return ()))
+{-# INLINE (=$$+) #-}
+
+-- | Continue processing after usage of '=$$+'. Connect a 'ResumableConduit' to
+-- a sink and return the output of the sink together with a new
+-- 'ResumableConduit'.
+--
+-- Since 1.0.17
+(=$$++) :: Monad m => ResumableConduit i m o -> Sink o m r -> Sink i m (ResumableConduit i m o, r)
+(=$$++) = connectResumeConduit
+{-# INLINE (=$$++) #-}
+
+-- | Complete processing of a 'ResumableConduit'. This will run the finalizer
+-- associated with the @ResumableConduit@. In order to guarantee process
+-- resource finalization, you /must/ use this operator after using '=$$+' and
+-- '=$$++'.
+--
+-- Since 1.0.17
+(=$$+-) :: Monad m => ResumableConduit i m o -> Sink o m r -> Sink i m r
+rsrc =$$+- sink = do
+    (ResumableConduit _ final, res) <- connectResumeConduit rsrc sink
+    lift final
+    return res
+{-# INLINE (=$$+-) #-}
+
+
+infixr 0 =$$+
+infixr 0 =$$++
+infixr 0 =$$+-
+
+-- | Provides an alternative @Applicative@ instance for @ConduitM@. In this instance,
+-- every incoming value is provided to all @ConduitM@s, and output is coalesced together.
+-- Leftovers from individual @ConduitM@s will be used within that component, and then discarded
+-- at the end of their computation. Output and finalizers will both be handled in a left-biased manner.
+--
+-- As an example, take the following program:
+--
+-- @
+-- main :: IO ()
+-- main = do
+--     let src = mapM_ yield [1..3 :: Int]
+--         conduit1 = CL.map (+1)
+--         conduit2 = CL.concatMap (replicate 2)
+--         conduit = getZipConduit $ ZipConduit conduit1 <* ZipConduit conduit2
+--         sink = CL.mapM_ print
+--     src $$ conduit =$ sink
+-- @
+--
+-- It will produce the output: 2, 1, 1, 3, 2, 2, 4, 3, 3
+--
+-- Since 1.0.17
+newtype ZipConduit i o m r = ZipConduit { getZipConduit :: ConduitM i o m r }
+    deriving Functor
+instance Monad m => Applicative (ZipConduit i o m) where
+    pure = ZipConduit . pure
+    ZipConduit left <*> ZipConduit right = ZipConduit (zipConduitApp left right)
+
+-- | Provide identical input to all of the @Conduit@s and combine their outputs
+-- into a single stream.
+--
+-- Implemented on top of @ZipConduit@, see that data type for more details.
+--
+-- Since 1.0.17
+sequenceConduits :: (Traversable f, Monad m) => f (ConduitM i o m r) -> ConduitM i o m (f r)
+sequenceConduits = getZipConduit . sequenceA . fmap ZipConduit
+
+-- | Fuse two @ConduitM@s together, and provide the return value of both. Note
+-- that this will force the entire upstream @ConduitM@ to be run to produce the
+-- result value, even if the downstream terminates early.
+--
+-- Since 1.1.5
+fuseBoth :: Monad m => ConduitM a b m r1 -> ConduitM b c m r2 -> ConduitM a c m (r1, r2)
+fuseBoth (ConduitM up) (ConduitM down) =
+    ConduitM $ pipeL up (withUpstream $ generalizeUpstream down)
+{-# INLINE fuseBoth #-}
+
+-- | Same as @fuseBoth@, but ignore the return value from the downstream
+-- @Conduit@. Same caveats of forced consumption apply.
+--
+-- Since 1.1.5
+fuseUpstream :: Monad m => ConduitM a b m r -> Conduit b m c -> ConduitM a c m r
+fuseUpstream up down = fmap fst (fuseBoth up down)
+{-# INLINE fuseUpstream #-}
