@@ -18,11 +18,13 @@ module Data.Conduit.Internal
     , Consumer
     , Conduit
     , ResumableSource (..)
+    , ResumableConduit (..)
       -- * Primitives
     , await
     , awaitE
     , awaitForever
     , yield
+    , yieldM
     , yieldOr
     , leftover
       -- * Finalization
@@ -33,10 +35,13 @@ module Data.Conduit.Internal
     , pipe
     , pipeL
     , connectResume
+    , connectResumeConduit
     , runPipe
     , injectLeftovers
     , (>+>)
     , (<+<)
+    , fuseLeftovers
+    , fuseReturnLeftovers
       -- * Generalizing
     , sourceToPipe
     , sinkToPipe
@@ -58,10 +63,16 @@ module Data.Conduit.Internal
     , sourceList
     , withUpstream
     , unwrapResumable
+    , unwrapResumableConduit
+    , newResumableSource
+    , newResumableConduit
     , Data.Conduit.Internal.enumFromTo
     , zipSinks
     , zipSources
     , zipSourcesApp
+    , zipConduitApp
+    , passthroughSink
+    , generalizeUpstream
     ) where
 
 import Control.Applicative (Applicative (..))
@@ -81,6 +92,9 @@ import Control.Monad.Trans.Resource
 import qualified GHC.Exts
 import qualified Data.IORef as I
 import Control.Monad.Morph (MFunctor (..))
+#if MIN_VERSION_exceptions(0, 6, 0)
+import qualified Control.Monad.Catch as Catch
+#endif
 
 -- | The underlying datatype for all the types in this package.  In has six
 -- type parameters:
@@ -148,10 +162,20 @@ instance MonadIO m => MonadIO (Pipe l i o u m) where
     liftIO = lift . liftIO
 
 instance MonadThrow m => MonadThrow (Pipe l i o u m) where
-    monadThrow = lift . monadThrow
+    throwM = lift . throwM
 
-instance MonadActive m => MonadActive (Pipe l i o u m) where
-    monadActive = lift monadActive
+#if MIN_VERSION_exceptions(0, 6, 0)
+instance Catch.MonadCatch m => Catch.MonadCatch (Pipe l i o u m) where
+    catch p0 onErr =
+        go p0
+      where
+        go (Done r) = Done r
+        go (PipeM mp) = PipeM $ Catch.catch (liftM go mp) (return . onErr)
+        go (Leftover p i) = Leftover (go p) i
+        go (NeedInput x y) = NeedInput (go . x) (go . y)
+        go (HaveOutput p c o) = HaveOutput (go p) c o
+    {-# INLINE catch #-}
+#endif
 
 instance Monad m => Monoid (Pipe l i o u m ()) where
     mempty = return ()
@@ -222,7 +246,11 @@ instance MonadError e m => MonadError e (Pipe l i o u m) where
 --
 -- Since 1.0.0
 newtype ConduitM i o m r = ConduitM { unConduitM :: Pipe i i o () m r }
-    deriving (Functor, Applicative, Monad, MonadIO, MonadTrans, MonadThrow, MonadActive, MonadResource, MFunctor)
+    deriving (Functor, Applicative, Monad, MonadIO, MonadTrans, MonadThrow, MFunctor
+#if MIN_VERSION_exceptions(0, 6, 0)
+    , Catch.MonadCatch
+#endif
+    )
 
 instance MonadReader r m => MonadReader r (ConduitM i o m) where
     ask = ConduitM ask
@@ -252,6 +280,9 @@ instance MonadError e m => MonadError e (ConduitM i o m) where
 instance MonadBase base m => MonadBase base (ConduitM i o m) where
     liftBase = lift . liftBase
 
+instance MonadResource m => MonadResource (ConduitM i o m) where
+    liftResourceT = lift . liftResourceT
+
 instance Monad m => Monoid (ConduitM i o m ()) where
     mempty = return ()
     mappend = (>>)
@@ -307,7 +338,7 @@ instance MFunctor ResumableSource where
 -- Since 0.5.0
 await :: Pipe l i o u m (Maybe i)
 await = NeedInput (Done . Just) (\_ -> Done Nothing)
-{-# RULES "await >>= maybe" forall x y. await >>= maybe x y = NeedInput y (const x) #-}
+{-# RULES "CI.await >>= maybe" forall x y. await >>= maybe x y = NeedInput y (const x) #-}
 {-# INLINE [1] await #-}
 
 -- | This is similar to @await@, but will return the upstream result value as
@@ -340,6 +371,10 @@ yield :: Monad m
 yield = HaveOutput (Done ()) (return ())
 {-# INLINE [1] yield #-}
 
+yieldM :: Monad m => m o -> Pipe l i o u m ()
+yieldM = PipeM . liftM (HaveOutput (Done ()) (return ()))
+{-# INLINE [1] yieldM #-}
+
 -- | Similar to @yield@, but additionally takes a finalizer to be run if the
 -- downstream @Pipe@ terminates.
 --
@@ -352,9 +387,11 @@ yieldOr o f = HaveOutput (Done ()) f o
 {-# INLINE [1] yieldOr #-}
 
 {-# RULES
-    "yield o >> p" forall o (p :: Pipe l i o u m r). yield o >> p = HaveOutput p (return ()) o
-  ; "mapM_ yield" mapM_ yield = sourceList
-  ; "yieldOr o c >> p" forall o c (p :: Pipe l i o u m r). yieldOr o c >> p = HaveOutput p c o #-}
+    "CI.yield o >> p" forall o (p :: Pipe l i o u m r). yield o >> p = HaveOutput p (return ()) o
+  ; "mapM_ CI.yield" mapM_ yield = sourceList
+  ; "CI.yieldOr o c >> p" forall o c (p :: Pipe l i o u m r). yieldOr o c >> p = HaveOutput p c o
+  ; "lift m >>= CI.yield" forall m. lift m >>= yield = yieldM m
+  #-}
 
 -- | Provide a single piece of leftover input to be consumed by the next pipe
 -- in the current monadic binding.
@@ -710,6 +747,13 @@ unwrapResumable (ResumableSource src final) = do
             x <- liftIO $ I.readIORef ref
             when x final
     return (liftIO (I.writeIORef ref False) >> src, final')
+
+-- | Turn a @Source@ into a @ResumableSource@ with no attached finalizer.
+--
+-- Since 1.1.4
+newResumableSource :: Monad m => Source m o -> ResumableSource m o
+newResumableSource s = ResumableSource s (return ())
+
 infixr 9 <+<
 infixl 9 >+>
 
@@ -913,3 +957,193 @@ zipSourcesApp (ConduitM left0) (ConduitM right0) =
     go (HaveOutput srcx closex x) (HaveOutput srcy closey y) = HaveOutput (go srcx srcy) (closex >> closey) (x y)
     go (NeedInput _ c) right = go (c ()) right
     go left (NeedInput _ c) = go left (c ())
+
+-- |
+--
+-- Since 1.0.17
+zipConduitApp
+    :: Monad m
+    => ConduitM i o m (x -> y)
+    -> ConduitM i o m x
+    -> ConduitM i o m y
+zipConduitApp (ConduitM left0) (ConduitM right0) =
+    ConduitM $ go (return ()) (return ()) (injectLeftovers left0) (injectLeftovers right0)
+  where
+    go _ _ (Done f) (Done x) = Done (f x)
+    go _ finalY (HaveOutput x finalX o) y = HaveOutput
+        (go finalX finalY x y)
+        (finalX >> finalY)
+        o
+    go finalX _ x (HaveOutput y finalY o) = HaveOutput
+        (go finalX finalY x y)
+        (finalX >> finalY)
+        o
+    go _ _ (Leftover _ i) _ = absurd i
+    go _ _ _ (Leftover _ i) = absurd i
+    go finalX finalY (PipeM mx) y = PipeM (flip (go finalX finalY) y `liftM` mx)
+    go finalX finalY x (PipeM my) = PipeM (go finalX finalY x `liftM` my)
+    go finalX finalY (NeedInput px cx) (NeedInput py cy) = NeedInput
+        (\i -> go finalX finalY (px i) (py i))
+        (\u -> go finalX finalY (cx u) (cy u))
+    go finalX finalY (NeedInput px cx) (Done y) = NeedInput
+        (\i -> go finalX finalY (px i) (Done y))
+        (\u -> go finalX finalY (cx u) (Done y))
+    go finalX finalY (Done x) (NeedInput py cy) = NeedInput
+        (\i -> go finalX finalY (Done x) (py i))
+        (\u -> go finalX finalY (Done x) (cy u))
+
+-- | Same as normal fusion (e.g. @=$=@), except instead of discarding leftovers
+-- from the downstream component, return them.
+--
+-- Since 1.0.17
+fuseReturnLeftovers :: Monad m
+                    => ConduitM a b m ()
+                    -> ConduitM b c m r
+                    -> ConduitM a c m (r, [b])
+fuseReturnLeftovers (ConduitM left0) (ConduitM right0) =
+    ConduitM $ goRight (return ()) [] left0 right0
+  where
+    goRight final bs left right =
+        case right of
+            HaveOutput p c o -> HaveOutput (recurse p) (c >> final) o
+            NeedInput rp rc  ->
+                case bs of
+                    [] -> goLeft rp rc final left
+                    b:bs' -> goRight final bs' left (rp b)
+            Done r2          -> PipeM (final >> return (Done (r2, bs)))
+            PipeM mp         -> PipeM (liftM recurse mp)
+            Leftover p b     -> goRight final (b:bs) left p
+      where
+        recurse = goRight final bs left
+
+    goLeft rp rc final left =
+        case left of
+            HaveOutput left' final' o -> goRight final' [] left' (rp o)
+            NeedInput left' lc        -> NeedInput (recurse . left') (recurse . lc)
+            Done r1                   -> goRight (return ()) [] (Done r1) (rc r1)
+            PipeM mp                  -> PipeM (liftM recurse mp)
+            Leftover left' i          -> Leftover (recurse left') i
+      where
+        recurse = goLeft rp rc final
+
+-- | Similar to @fuseReturnLeftovers@, but use the provided function to convert
+-- downstream leftovers to upstream leftovers.
+--
+-- Since 1.0.17
+fuseLeftovers
+    :: Monad m
+    => ([b] -> [a])
+    -> ConduitM a b m ()
+    -> ConduitM b c m r
+    -> ConduitM a c m r
+fuseLeftovers f left right = do
+    (r, bs) <- fuseReturnLeftovers left right
+    ConduitM $ mapM_ leftover $ reverse $ f bs
+    return r
+
+-- | A generalization of 'ResumableSource'. Allows to resume an arbitrary
+-- conduit, keeping its state and using it later (or finalizing it).
+--
+-- Since 1.0.17
+data ResumableConduit i m o =
+    ResumableConduit (Conduit i m o) (m ())
+
+-- | Connect a 'ResumableConduit' to a sink and return the output of the sink
+-- together with a new 'ResumableConduit'.
+--
+-- Since 1.0.17
+connectResumeConduit
+    :: Monad m
+    => ResumableConduit i m o
+    -> Sink o m r
+    -> Sink i m (ResumableConduit i m o, r)
+connectResumeConduit (ResumableConduit (ConduitM left0) leftFinal0) (ConduitM right0) =
+    ConduitM $ goRight leftFinal0 left0 right0
+  where
+    goRight leftFinal left right =
+        case right of
+            HaveOutput _ _ o -> absurd o
+            NeedInput rp rc -> goLeft rp rc leftFinal left
+            Done r2 -> Done (ResumableConduit (ConduitM left) leftFinal, r2)
+            PipeM mp -> PipeM (liftM (goRight leftFinal left) mp)
+            Leftover p i -> goRight leftFinal (HaveOutput left leftFinal i) p
+
+    goLeft rp rc leftFinal left =
+        case left of
+            HaveOutput left' leftFinal' o -> goRight leftFinal' left' (rp o)
+            NeedInput left' lc -> NeedInput (recurse . left') (recurse . lc)
+            Done () -> goRight (return ()) (Done ()) (rc ())
+            PipeM mp -> PipeM (liftM recurse mp)
+            Leftover left' i -> Leftover (recurse left') i -- recurse p
+      where
+        recurse = goLeft rp rc leftFinal
+
+-- | Unwraps a @ResumableConduit@ into a @Conduit@ and a finalizer.
+--
+-- Since 'unwrapResumable' for more information.
+--
+-- Since 1.0.17
+unwrapResumableConduit :: MonadIO m => ResumableConduit i m o -> m (Conduit i m o, m ())
+unwrapResumableConduit (ResumableConduit src final) = do
+    ref <- liftIO $ I.newIORef True
+    let final' = do
+            x <- liftIO $ I.readIORef ref
+            when x final
+    return (liftIO (I.writeIORef ref False) >> src, final')
+
+-- | Turn a @Conduit@ into a @ResumableConduit@ with no attached finalizer.
+--
+-- Since 1.1.4
+newResumableConduit :: Monad m => Conduit i m o -> ResumableConduit i m o
+newResumableConduit c = ResumableConduit c (return ())
+
+-- | Turn a @Sink@ into a @Conduit@ in the following way:
+--
+-- * All input passed to the @Sink@ is yielded downstream.
+--
+-- * When the @Sink@ finishes processing, the result is passed to the provided to the finalizer function.
+--
+-- Note that the @Sink@ will stop receiving input as soon as the downstream it
+-- is connected to shuts down.
+--
+-- An example usage would be to write the result of a @Sink@ to some mutable
+-- variable while allowing other processing to continue.
+--
+-- Since 1.1.0
+passthroughSink :: Monad m
+                => Sink i m r
+                -> (r -> m ()) -- ^ finalizer
+                -> Conduit i m i
+passthroughSink (ConduitM sink0) final =
+    ConduitM $ go [] sink0
+  where
+    go _ (Done r) = do
+        lift $ final r
+        awaitForever yield
+    go is (Leftover sink i) = go (i:is) sink
+    go _ (HaveOutput _ _ o) = absurd o
+    go is (PipeM mx) = do
+        x <- lift mx
+        go is x
+    go (i:is) (NeedInput next _) = go is (next i)
+    go [] (NeedInput next done) = do
+        mx <- await
+        case mx of
+            Nothing -> go [] (done ())
+            Just x -> do
+                yield x
+                go [] (next x)
+
+-- | Generalize the upstream return value for a @Pipe@ from unit to any type.
+--
+-- Since 1.1.5
+generalizeUpstream :: Monad m => Pipe l i o () m r -> Pipe l i o u m r
+generalizeUpstream =
+    go
+  where
+    go (HaveOutput p f o) = HaveOutput (go p) f o
+    go (NeedInput x y) = NeedInput (go . x) (\_ -> go (y ()))
+    go (Done r) = Done r
+    go (PipeM mp) = PipeM (liftM go mp)
+    go (Leftover p l) = Leftover (go p) l
+{-# INLINE generalizeUpstream #-}
