@@ -1,32 +1,22 @@
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE DeriveDataTypeable #-}
 import Test.Hspec
 import Test.Hspec.QuickCheck (prop)
 import Test.QuickCheck.Monadic (assert, monadicIO, run)
 
-import Control.Exception (IOException)
 import qualified Data.Conduit as C
 import qualified Data.Conduit.Lift as C
-import qualified Data.Conduit.Util as C
 import qualified Data.Conduit.Internal as CI
 import qualified Data.Conduit.List as CL
-import qualified Data.Conduit.Lazy as CLazy
-import qualified Data.Conduit.Binary as CB
-import qualified Data.Conduit.Text as CT
-import Data.Conduit (runResourceT)
-import Data.Maybe   (fromMaybe,catMaybes)
+import Data.Typeable (Typeable)
+import Control.Monad.Trans.Resource as C (runResourceT)
+import Data.Maybe   (fromMaybe,catMaybes,fromJust)
 import qualified Data.List as DL
 import Control.Monad.ST (runST)
 import Data.Monoid
-import qualified Data.ByteString as S
-import qualified Data.ByteString.Char8 as S8
 import qualified Data.IORef as I
-import qualified Data.ByteString.Lazy as L
-import Data.ByteString.Lazy.Char8 ()
-import qualified Data.Text as T
-import qualified Data.Text.Lazy as TL
-import qualified Data.Text.Lazy.Encoding as TLE
-import Control.Monad.Trans.Resource (runExceptionT, runExceptionT_, allocate, resourceForkIO)
+import Control.Monad.Trans.Resource (allocate, resourceForkIO)
 import Control.Concurrent (threadDelay, killThread)
 import Control.Monad.IO.Class (MonadIO, liftIO)
 import Control.Monad.Trans.Class (lift)
@@ -34,11 +24,14 @@ import Control.Monad.Trans.Writer (execWriter, tell, runWriterT)
 import Control.Monad.Trans.State (evalStateT, get, put, modify)
 import Control.Monad.Trans.Maybe (MaybeT (..))
 import Control.Applicative (pure, (<$>), (<*>))
+import qualified Control.Monad.Catch as Catch
 import Data.Functor.Identity (Identity,runIdentity)
 import Control.Monad (forever, void)
 import Data.Void (Void)
 import qualified Control.Concurrent.MVar as M
 import Control.Monad.Error (catchError, throwError, Error)
+import qualified Data.Map as Map
+import qualified Data.Conduit.Extra.ZipConduitSpec as ZipConduit
 
 (@=?) :: (Eq a, Show a) => a -> a -> IO ()
 (@=?) = flip shouldBe
@@ -81,7 +74,7 @@ main = hspec $ do
         prop "Maybe" $
             equivToList (catMaybes . map f :: [Int]->[Int]) (CL.mapFoldable f)
 
-    prop "scanl" $ equivToList (tail . scanl (+) 0 :: [Int]->[Int]) (CL.scanl (\a s -> (a+s,a+s)) 0)
+    prop "scan" $ equivToList (tail . scanl (+) 0 :: [Int]->[Int]) (void $ CL.scan (+) 0)
 
     -- mapFoldableM and scanlM are fully polymorphic in type of monad
     -- so it suffice to check only with Identity.
@@ -92,7 +85,7 @@ main = hspec $ do
         prop "Maybe" $
             equivToList (catMaybes . map f :: [Int]->[Int]) (CL.mapFoldableM (return . f))
 
-    prop "scanl" $ equivToList (tail . scanl (+) 0 :: [Int]->[Int]) (CL.scanlM (\a s -> return (a+s,a+s)) 0)
+    prop "scanM" $ equivToList (tail . scanl (+) 0) (void $ CL.scanM (\a s -> return $ a + s) (0 :: Int))
 
     describe "ResourceT" $ do
         it "resourceForkIO" $ do
@@ -153,6 +146,15 @@ main = hspec $ do
             let y = DL.unfoldr f seed
             x `shouldBe` y
 
+    describe "unfoldM" $ do
+        it "works" $ do
+            let f 0 = Nothing
+                f i = Just (show i, i - 1)
+                seed = 10 :: Int
+            x <- CL.unfoldM (return . f) seed C.$$ CL.consume
+            let y = DL.unfoldr f seed
+            x `shouldBe` y
+
     describe "Monoid instance for Source" $ do
         it "mappend" $ do
             x <- runResourceT $ (CL.sourceList [1..5 :: Int] `mappend` CL.sourceList [6..10]) C.$$ CL.fold (+) 0
@@ -165,47 +167,20 @@ main = hspec $ do
                 ] C.$$ CL.fold (+) 0
             x `shouldBe` sum [1..20]
 
-    describe "file access" $ do
-        it "read" $ do
-            bs <- S.readFile "conduit.cabal"
-            bss <- runResourceT $ CB.sourceFile "conduit.cabal" C.$$ CL.consume
-            bs @=? S.concat bss
-
-        it "read range" $ do
-            S.writeFile "tmp" "0123456789"
-            bss <- runResourceT $ CB.sourceFileRange "tmp" (Just 2) (Just 3) C.$$ CL.consume
-            S.concat bss `shouldBe` "234"
-
-        it "write" $ do
-            runResourceT $ CB.sourceFile "conduit.cabal" C.$$ CB.sinkFile "tmp"
-            bs1 <- S.readFile "conduit.cabal"
-            bs2 <- S.readFile "tmp"
-            bs1 @=? bs2
-
-        it "conduit" $ do
-            runResourceT $ CB.sourceFile "conduit.cabal"
-                C.$= CB.conduitFile "tmp"
-                C.$$ CB.sinkFile "tmp2"
-            bs1 <- S.readFile "conduit.cabal"
-            bs2 <- S.readFile "tmp"
-            bs3 <- S.readFile "tmp2"
-            bs1 @=? bs2
-            bs1 @=? bs3
-
     describe "zipping" $ do
         it "zipping two small lists" $ do
-            res <- runResourceT $ C.zipSources (CL.sourceList [1..10]) (CL.sourceList [11..12]) C.$$ CL.consume
+            res <- runResourceT $ CI.zipSources (CL.sourceList [1..10]) (CL.sourceList [11..12]) C.$$ CL.consume
             res @=? zip [1..10 :: Int] [11..12 :: Int]
 
     describe "zipping sinks" $ do
         it "take all" $ do
-            res <- runResourceT $ CL.sourceList [1..10] C.$$ C.zipSinks CL.consume CL.consume
+            res <- runResourceT $ CL.sourceList [1..10] C.$$ CI.zipSinks CL.consume CL.consume
             res @=? ([1..10 :: Int], [1..10 :: Int])
         it "take fewer on left" $ do
-            res <- runResourceT $ CL.sourceList [1..10] C.$$ C.zipSinks (CL.take 4) CL.consume
+            res <- runResourceT $ CL.sourceList [1..10] C.$$ CI.zipSinks (CL.take 4) CL.consume
             res @=? ([1..4 :: Int], [1..10 :: Int])
         it "take fewer on right" $ do
-            res <- runResourceT $ CL.sourceList [1..10] C.$$ C.zipSinks CL.consume (CL.take 4)
+            res <- runResourceT $ CL.sourceList [1..10] C.$$ CI.zipSinks CL.consume (CL.take 4)
             res @=? ([1..10 :: Int], [1..4 :: Int])
 
     describe "Monad instance for Sink" $ do
@@ -338,33 +313,6 @@ main = hspec $ do
                 CL.consume
             x `shouldBe` [6..10]
 
-    describe "lazy" $ do
-        it' "works inside a ResourceT" $ runResourceT $ do
-            counter <- liftIO $ I.newIORef 0
-            let incr i = do
-                    istate <- liftIO $ I.newIORef $ Just (i :: Int)
-                    let loop = do
-                            res <- liftIO $ I.atomicModifyIORef istate ((,) Nothing)
-                            case res of
-                                Nothing -> return ()
-                                Just x -> do
-                                    count <- liftIO $ I.atomicModifyIORef counter
-                                        (\j -> (j + 1, j + 1))
-                                    liftIO $ count `shouldBe` i
-                                    C.yield x
-                                    loop
-                    loop
-            nums <- CLazy.lazyConsume $ mconcat $ map incr [1..10]
-            liftIO $ nums `shouldBe` [1..10]
-
-        it' "returns nothing outside ResourceT" $ do
-            bss <- runResourceT $ CLazy.lazyConsume $ CB.sourceFile "test/main.hs"
-            bss `shouldBe` []
-
-        it' "works with pure sources" $ do
-            nums <- CLazy.lazyConsume $ forever $ C.yield 1
-            take 100 nums `shouldBe` replicate 100 (1 :: Int)
-
     describe "sequence" $ do
         it "simple sink" $ do
             let sumSink = do
@@ -400,88 +348,6 @@ main = hspec $ do
                 return (a, b)
             (a, b) `shouldBe` (Just 1, [1..10])
 
-    describe "text" $ do
-        let go enc tenc tdec cenc = do
-                prop (enc ++ " single chunk") $ \chars -> runST $ runExceptionT_ $ do
-                    let tl = TL.pack chars
-                        lbs = tenc tl
-                        src = CL.sourceList $ L.toChunks lbs
-                    ts <- src C.$= CT.decode cenc C.$$ CL.consume
-                    return $ TL.fromChunks ts == tl
-                prop (enc ++ " many chunks") $ \chars -> runIdentity $ runExceptionT_ $ do
-                    let tl = TL.pack chars
-                        lbs = tenc tl
-                        src = mconcat $ map (CL.sourceList . return . S.singleton) $ L.unpack lbs
-                        
-                    ts <- src C.$= CT.decode cenc C.$$ CL.consume
-                    return $ TL.fromChunks ts == tl
-
-                -- Check whether raw bytes are decoded correctly, in
-                -- particular that Text decoding produces an error if
-                -- and only if Conduit does.
-                prop (enc ++ " raw bytes") $ \bytes ->
-                    let lbs = L.pack bytes
-                        src = CL.sourceList $ L.toChunks lbs
-                        etl = C.runException $ src C.$= CT.decode cenc C.$$ CL.consume
-                        tl' = tdec lbs
-                    in  case etl of
-                          (Left _) -> (return $! TL.toStrict tl') `shouldThrow` anyException
-                          (Right tl) -> TL.fromChunks tl `shouldBe` tl'
-                prop (enc ++ " encoding") $ \chars -> runIdentity $ runExceptionT_ $ do
-                    let tss = map T.pack chars
-                        lbs = tenc $ TL.fromChunks tss
-                        src = mconcat $ map (CL.sourceList . return) tss
-                    bss <- src C.$= CT.encode cenc C.$$ CL.consume
-                    return $ L.fromChunks bss == lbs
-        go "utf8" TLE.encodeUtf8 TLE.decodeUtf8 CT.utf8
-        go "utf16_le" TLE.encodeUtf16LE TLE.decodeUtf16LE CT.utf16_le
-        go "utf16_be" TLE.encodeUtf16BE TLE.decodeUtf16BE CT.utf16_be
-        go "utf32_le" TLE.encodeUtf32LE TLE.decodeUtf32LE CT.utf32_le
-        go "utf32_be" TLE.encodeUtf32BE TLE.decodeUtf32BE CT.utf32_be
-
-    describe "text lines" $ do
-        it "works across split lines" $
-            (CL.sourceList [T.pack "abc", T.pack "d\nef"] C.$= CT.lines C.$$ CL.consume) ==
-                [[T.pack "abcd", T.pack "ef"]]
-        it "works with multiple lines in an item" $
-            (CL.sourceList [T.pack "ab\ncd\ne"] C.$= CT.lines C.$$ CL.consume) ==
-                [[T.pack "ab", T.pack "cd", T.pack "e"]]
-        it "works with ending on a newline" $
-            (CL.sourceList [T.pack "ab\n"] C.$= CT.lines C.$$ CL.consume) ==
-                [[T.pack "ab"]]
-        it "works with ending a middle item on a newline" $
-            (CL.sourceList [T.pack "ab\n", T.pack "cd\ne"] C.$= CT.lines C.$$ CL.consume) ==
-                [[T.pack "ab", T.pack "cd", T.pack "e"]]
-        it "is not too eager" $ do
-            x <- CL.sourceList ["foobarbaz", error "ignore me"] C.$$ CT.decode CT.utf8 C.=$ CL.head
-            x `shouldBe` Just "foobarbaz"
-
-    describe "text lines bounded" $ do
-        it "works across split lines" $
-            (CL.sourceList [T.pack "abc", T.pack "d\nef"] C.$= CT.linesBounded 80 C.$$ CL.consume) ==
-                [[T.pack "abcd", T.pack "ef"]]
-        it "works with multiple lines in an item" $
-            (CL.sourceList [T.pack "ab\ncd\ne"] C.$= CT.linesBounded 80 C.$$ CL.consume) ==
-                [[T.pack "ab", T.pack "cd", T.pack "e"]]
-        it "works with ending on a newline" $
-            (CL.sourceList [T.pack "ab\n"] C.$= CT.linesBounded 80 C.$$ CL.consume) ==
-                [[T.pack "ab"]]
-        it "works with ending a middle item on a newline" $
-            (CL.sourceList [T.pack "ab\n", T.pack "cd\ne"] C.$= CT.linesBounded 80 C.$$ CL.consume) ==
-                [[T.pack "ab", T.pack "cd", T.pack "e"]]
-        it "is not too eager" $ do
-            x <- CL.sourceList ["foobarbaz", error "ignore me"] C.$$ CT.decode CT.utf8 C.=$ CL.head
-            x `shouldBe` Just "foobarbaz"
-        it "throws an exception when lines are too long" $ do
-            x <- C.runExceptionT $ CL.sourceList ["hello\nworld"] C.$$ CT.linesBounded 4 C.=$ CL.consume
-            show x `shouldBe` show (Left $ CT.LengthExceeded 4 :: Either CT.TextException ())
-
-    describe "binary isolate" $ do
-        it "works" $ do
-            bss <- runResourceT $ CL.sourceList (replicate 1000 "X")
-                           C.$= CB.isolate 6
-                           C.$$ CL.consume
-            S.concat bss `shouldBe` "XXXXXX"
     describe "unbuffering" $ do
         it "works" $ do
             x <- runResourceT $ do
@@ -525,97 +391,6 @@ main = hspec $ do
             x `shouldBe` sum [1..10]
 
 
-    describe "properly using binary file reading" $ do
-        it "sourceFile" $ do
-            x <- runResourceT $ CB.sourceFile "test/random" C.$$ CL.consume
-            lbs <- L.readFile "test/random"
-            L.fromChunks x `shouldBe` lbs
-
-    describe "binary head" $ do
-        let go lbs = do
-                x <- CB.head
-                case (x, L.uncons lbs) of
-                    (Nothing, Nothing) -> return True
-                    (Just y, Just (z, lbs'))
-                        | y == z -> go lbs'
-                    _ -> return False
-
-        prop "works" $ \bss' ->
-            let bss = map S.pack bss'
-             in runIdentity $
-                CL.sourceList bss C.$$ go (L.fromChunks bss)
-    describe "binary takeWhile" $ do
-        prop "works" $ \bss' ->
-            let bss = map S.pack bss'
-             in runIdentity $ do
-                bss2 <- CL.sourceList bss C.$$ CB.takeWhile (>= 5) C.=$ CL.consume
-                return $ L.fromChunks bss2 == L.takeWhile (>= 5) (L.fromChunks bss)
-        prop "leftovers present" $ \bss' ->
-            let bss = map S.pack bss'
-             in runIdentity $ do
-                result <- CL.sourceList bss C.$$ do
-                    x <- CB.takeWhile (>= 5) C.=$ CL.consume
-                    y <- CL.consume
-                    return (S.concat x, S.concat y)
-                let expected = S.span (>= 5) $ S.concat bss
-                if result == expected
-                    then return True
-                    else error $ show (S.concat bss, result, expected)
-
-    describe "binary dropWhile" $ do
-        prop "works" $ \bss' ->
-            let bss = map S.pack bss'
-             in runIdentity $ do
-                bss2 <- CL.sourceList bss C.$$ do
-                    CB.dropWhile (< 5)
-                    CL.consume
-                return $ L.fromChunks bss2 == L.dropWhile (< 5) (L.fromChunks bss)
-
-    describe "binary take" $ do
-      let go n l = CL.sourceList l C.$$ do
-          a <- CB.take n
-          b <- CL.consume
-          return (a, b)
-
-      -- Taking nothing should result in an empty Bytestring
-      it "nothing" $ do
-        (a, b) <- runResourceT $ go 0 ["abc", "defg"]
-        a              `shouldBe` L.empty
-        L.fromChunks b `shouldBe` "abcdefg"
-
-      it "normal" $ do
-        (a, b) <- runResourceT $ go 4 ["abc", "defg"]
-        a              `shouldBe` "abcd"
-        L.fromChunks b `shouldBe` "efg"
-
-      -- Taking exactly the data that is available should result in no
-      -- leftover.
-      it "all" $ do
-        (a, b) <- runResourceT $ go 7 ["abc", "defg"]
-        a `shouldBe` "abcdefg"
-        b `shouldBe` []
-
-      -- Take as much as possible.
-      it "more" $ do
-        (a, b) <- runResourceT $ go 10 ["abc", "defg"]
-        a `shouldBe` "abcdefg"
-        b `shouldBe` []
-
-    describe "normalFuseLeft" $ do
-        it "does not double close conduit" $ do
-            x <- runResourceT $ do
-                let src = CL.sourceList ["foobarbazbin"]
-                src C.$= CB.isolate 10 C.$$ CL.head
-            x `shouldBe` Just "foobarbazb"
-
-    describe "binary" $ do
-        prop "lines" $ \bss' -> runIdentity $ do
-            let bss = map S.pack bss'
-                bs = S.concat bss
-                src = CL.sourceList bss
-            res <- src C.$$ CB.lines C.=$ CL.consume
-            return $ S8.lines bs == res
-
     describe "termination" $ do
         it "terminates early" $ do
             let src = forever $ C.yield ()
@@ -889,12 +664,6 @@ main = hspec $ do
             x <- runWriterT $ source C.$$ C.transPipe (`evalStateT` 1) replaceNum1 C.=$ CL.consume
             y <- runWriterT $ source C.$$ C.transPipe (`evalStateT` 1) replaceNum2 C.=$ CL.consume
             x `shouldBe` y
-    describe "text decode" $ do
-        it' "doesn't throw runtime exceptions" $ do
-            let x = runIdentity $ runExceptionT $ C.yield "\x89\x243" C.$$ CT.decode CT.utf8 C.=$ CL.consume
-            case x of
-                Left _ -> return ()
-                Right t -> error $ "This should have failed: " ++ show t
     describe "iterM" $ do
         prop "behavior" $ \l -> monadicIO $ do
             let counter ref = CL.iterM (const $ liftIO $ M.modifyMVar_ ref (\i -> return $! i + 1))
@@ -929,24 +698,6 @@ main = hspec $ do
             res <- C.yield 10 C.$$ C.awaitForever (C.toProducer . src) C.=$ (C.toConsumer sink >>= C.yield) C.=$ C.await
             res `shouldBe` Just (sum [1..10])
 
-    describe "sinkCacheLength" $ do
-        it' "works" $ C.runResourceT $ do
-            lbs <- liftIO $ L.readFile "test/main.hs"
-            (len, src) <- CB.sourceLbs lbs C.$$ CB.sinkCacheLength
-            lbs' <- src C.$$ CB.sinkLbs
-            liftIO $ do
-                fromIntegral len `shouldBe` L.length lbs
-                lbs' `shouldBe` lbs
-                fromIntegral len `shouldBe` L.length lbs'
-
-    describe "Data.Conduit.Binary.mapM_" $ do
-        prop "telling works" $ \bytes ->
-            let lbs = L.pack bytes
-                src = CB.sourceLbs lbs
-                sink = CB.mapM_ (tell . return . S.singleton)
-                bss = execWriter $ src C.$$ sink
-             in L.fromChunks bss == lbs
-
     describe "passthroughSink" $ do
         it "works" $ do
             ref <- I.newIORef (-1)
@@ -1055,8 +806,8 @@ main = hspec $ do
                 test1L = testPipe $ (p1 <+< p2) <+< p3
                 test1R = testPipe $ p1 <+< (p2 <+< p3)
 
-                test2L = testPipe $ (p2 <+< p1) <+< p3
-                test2R = testPipe $ p2 <+< (p1 <+< p3)
+                _test2L = testPipe $ (p2 <+< p1) <+< p3
+                _test2R = testPipe $ p2 <+< (p1 <+< p3)
 
                 test3L = testPipe $ (p2 <+< p3) <+< p1
                 test3R = testPipe $ p2 <+< (p3 <+< p1)
@@ -1081,13 +832,13 @@ main = hspec $ do
 
     describe "Data.Conduit.Lift" $ do
         it "execStateC" $ do
-            let sink = C.execStateC 0 $ CL.mapM_ $ modify . (+)
+            let sink = C.execStateLC 0 $ CL.mapM_ $ modify . (+)
                 src = mapM_ C.yield [1..10 :: Int]
             res <- src C.$$ sink
             res `shouldBe` sum [1..10]
 
         it "execWriterC" $ do
-            let sink = C.execWriterC $ CL.mapM_ $ tell . return
+            let sink = C.execWriterLC $ CL.mapM_ $ tell . return
                 src = mapM_ C.yield [1..10 :: Int]
             res <- src C.$$ sink
             res `shouldBe` [1..10]
@@ -1108,37 +859,66 @@ main = hspec $ do
             res <- src C.$$ sink
             res `shouldBe` [1 :: Int]
 
-    describe "exception handling" $ do
-        it "catchC" $ do
-            ref <- I.newIORef 0
+    describe "sequenceSources" $ do
+        it "works" $ do
+            let src1 = mapM_ C.yield [1, 2, 3 :: Int]
+                src2 = mapM_ C.yield [3, 2, 1]
+                src3 = mapM_ C.yield $ repeat 2
+                srcs = C.sequenceSources $ Map.fromList
+                    [ (1 :: Int, src1)
+                    , (2, src2)
+                    , (3, src3)
+                    ]
+            res <- srcs C.$$ CL.consume
+            res `shouldBe`
+                [ Map.fromList [(1, 1), (2, 3), (3, 2)]
+                , Map.fromList [(1, 2), (2, 2), (3, 2)]
+                , Map.fromList [(1, 3), (2, 1), (3, 2)]
+                ]
+    describe "zipSink" $ do
+        it "zip equal-sized" $ do
+            x <- runResourceT $
+                    CL.sourceList [1..100] C.$$
+                    C.sequenceSinks [ CL.fold (+) 0,
+                                   (`mod` 101) <$> CL.fold (*) 1 ]
+            x `shouldBe` [5050, 100 :: Integer]
+
+        it "zip distinct sizes" $ do
+            let sink = C.getZipSink $
+                        (*) <$> C.ZipSink (CL.fold (+) 0)
+                            <*> C.ZipSink (Data.Maybe.fromJust <$> C.await)
+            x <- C.runResourceT $ CL.sourceList [100,99..1] C.$$ sink
+            x `shouldBe` (505000 :: Integer)
+
+    describe "upstream results" $ do
+        it "fuseBoth" $ do
+            let upstream = do
+                    C.yield ("hello" :: String)
+                    CL.isolate 5 C.=$= CL.fold (+) 0
+                downstream = C.fuseBoth upstream CL.consume
+            res <- CL.sourceList [1..10 :: Int] C.$$ do
+                (x, y) <- downstream
+                z <- CL.consume
+                return (x, y, z)
+            res `shouldBe` (sum [1..5], ["hello"], [6..10])
+
+    describe "catching exceptions" $ do
+        it "works" $ do
             let src = do
-                    C.catchC (CB.sourceFile "some-file-that-does-not-exist") onErr
-                    C.handleC onErr $ CB.sourceFile "conduit.cabal"
-                onErr :: MonadIO m => IOException -> m ()
-                onErr _ = liftIO $ I.modifyIORef ref (+ 1)
-            contents <- L.readFile "conduit.cabal"
-            res <- C.runResourceT $ src C.$$ CB.sinkLbs
-            res `shouldBe` contents
-            errCount <- I.readIORef ref
-            errCount `shouldBe` (1 :: Int)
-        it "tryC" $ do
-            ref <- I.newIORef undefined
-            let src = do
-                    res1 <- C.tryC $ CB.sourceFile "some-file-that-does-not-exist"
-                    res2 <- C.tryC $ CB.sourceFile "conduit.cabal"
-                    liftIO $ I.writeIORef ref (res1, res2)
-            contents <- L.readFile "conduit.cabal"
-            res <- C.runResourceT $ src C.$$ CB.sinkLbs
-            res `shouldBe` contents
-            exc <- I.readIORef ref
-            case exc :: (Either IOException (), Either IOException ()) of
-                (Left _, Right ()) ->
-                    return ()
-                _ -> error $ show exc
+                    C.yield 1
+                    () <- Catch.throwM DummyError
+                    C.yield 2
+                src' = do
+                    Catch.catch src (\DummyError -> C.yield (3 :: Int))
+            res <- src' C.$$ CL.consume
+            res `shouldBe` [1, 3]
+
+    ZipConduit.spec
 
 it' :: String -> IO () -> Spec
 it' = it
 
 data DummyError = DummyError
-    deriving (Show, Eq)
+    deriving (Show, Eq, Typeable)
 instance Error DummyError
+instance Catch.Exception DummyError
